import random
from typing import Literal, Tuple, Type

from ai_hexagon.test import BaseTest
from ai_hexagon.model import Model


class HashMap(BaseTest):
    __title__ = "Hash Map"
    __description__ = (
        "Tests the model's capacity to memorize key-value pairs from the training data."
    )
    name: Literal["hash_map"] = "hash_map"

    key_length: int = 8
    value_length: int = 64
    num_pairs_range: Tuple[int, int] = (32, 65536)
    vocab_size: int = 1024

    def evalulate(self, model: Type[Model]) -> float:
        # TODO: implement the testing
        return random.random()



from abc import ABC, abstractmethod
from typing import ClassVar, Optional, Type
import jax
from flax.typing import Array
from pydantic import BaseModel

from ai_hexagon.model import Model


class BaseTest(ABC, BaseModel):
    __title__: ClassVar[str] = ""
    __description__: ClassVar[Optional[str]] = None

    __key__: Optional[Array] = None

    name: str
    seed: int = 0

    def __hash__(self):
        return self.model_dump_json().__hash__()

    def __eq__(self, other):
        return self.model_dump_json() == other.model_dump_json()

    def model_post_init(self, __context):
        self.__key__ = jax.random.PRNGKey(self.seed)
        return super().model_post_init(__context)

    @abstractmethod
    def evalulate(self, model: Type[Model]) -> float: ...

    @classmethod
    def get_test_name(cls):
        return cls.model_fields["name"].default

    @property
    def key(self):
        self.__key__, subkey = jax.random.split(self.__key__)
        return subkey



import flax.linen as nn
import jax.numpy as jnp
from flax.typing import Array


class GroupedQueryAttention(nn.Module):
    dims: int
    q_heads: int
    k_heads: int

    @nn.compact
    def __call__(self, q: Array, k: Array, v: Array, mask: Array = None) -> Array:
        assert (
            self.q_heads % self.k_heads == 0
        ), "num of kv heads must by divisible by num of q heads"
        assert self.dims % self.q_heads == 0, "dims must be devisible by num of q heads"

        in_dim = q.shape[-1]

        head_dim = self.dims // self.q_heads

        qx = nn.Dense(self.q_heads * head_dim)(q)
        qx = qx.reshape((*qx.shape[:-1], self.q_heads, head_dim))

        kx = nn.Dense(self.k_heads * head_dim)(k)
        kx = kx.reshape((*kx.shape[:-1], self.k_heads, head_dim))

        vx = nn.Dense(self.k_heads * head_dim)(v)
        vx = vx.reshape((*vx.shape[:-1], self.k_heads, head_dim))

        kx = kx.repeat(self.q_heads // self.k_heads, axis=-2)
        vx = vx.repeat(self.q_heads // self.k_heads, axis=-2)

        qx = jnp.einsum("...ijk->...jik", qx)
        kx = jnp.einsum("...ijk->...jik", kx)
        vx = jnp.einsum("...ijk->...jik", vx)
        scores = jnp.einsum("...jk,...ik->...ji", qx, kx)
        scores /= jnp.sqrt(head_dim)
        if mask is not None:
            scores -= 1 / mask + 1
        scores = nn.softmax(scores)

        out = jnp.einsum("...jk,...ki->...ji", scores, vx)
        out = jnp.einsum("...ijk->...jik", out)
        out = out.reshape((*out.shape[:-2], -1))
        out = nn.Dense(in_dim)(out)

        return out



import flax.linen as nn


class MLP(nn.Module):
    dims: int
    layers: int = 1

    @nn.compact
    def __call__(self, x):
        in_dims = x.shape[-1]
        for _ in range(self.layers):
            x = nn.Dense(self.dims)(x)
            x = nn.silu(x)
        x = nn.Dense(in_dims)(x)
        return x



import operator
from typing import ClassVar, Dict, List, Optional
import inflection
import jax
import jax.numpy as jnp
import flax.linen as nn
from flax.typing import Array, FrozenVariableDict
from flax.training.train_state import TrainState
import optax
from pydantic import BaseModel


class ModelStats(BaseModel):
    size: int
    size_doubling_rate: float
    size_big_o: str
    flops: int
    flops_doubling_rate: float
    flops_big_o: str


def _compute_doubling_rate(x: Dict):
    n = jnp.array(list(x.keys()))
    fn = jnp.array(list(x.values()))
    slope = jnp.polyfit(jnp.log2(n), jnp.log2(fn), 1)[0]
    if abs(slope) < 1e-3:
        return 0.0
    return float(slope)


def _fit_big_o(x: Dict) -> str:
    n = jnp.array(list(x.keys()))  # noqa F841
    fn = jnp.array(list(x.values()))  # noqa F841
    # TODO
    return "???"


class Model(nn.Module):
    vocab_size: int

    __title__: ClassVar[Optional[str]] = None
    __authors__: ClassVar[Optional[List[str]]] = None
    __paper__: ClassVar[Optional[str]] = None

    @classmethod
    def get_model_title(cls) -> str:
        return cls.__title__ or inflection.titleize(cls.__name__)

    @classmethod
    def compute_stats(
        cls, vocab_size: int, sequence_length: int, sequence_lengths: List[int]
    ) -> ModelStats:
        lenghts = set(sequence_lengths) | {sequence_length}
        sizes: Dict[int, int] = {}
        flops: Dict[float, float] = {}
        for length in lenghts:
            model = cls(vocab_size=vocab_size)
            x = jnp.zeros((1, length), dtype=jnp.uint32)
            variables = model.init(jax.random.PRNGKey(0), x)
            params = variables["params"]
            sizes[length] = jax.tree.reduce(
                operator.add, jax.tree.map(lambda x: x.nbytes, params)
            )
            compiled = jax.jit(model.apply).lower(variables, x).compile()
            flops[length] = float(compiled.cost_analysis()[0]["flops"])

        return ModelStats(
            size=sizes[sequence_length],
            size_doubling_rate=_compute_doubling_rate(sizes),
            size_big_o=_fit_big_o(sizes),
            flops=flops[sequence_length],
            flops_doubling_rate=_compute_doubling_rate(flops),
            flops_big_o=_fit_big_o(sizes),
        )

    def init_train_state(self, x: Array, key: Array) -> TrainState:
        variables = self.init(key, x)
        state = TrainState.create(
            apply_fn=self.apply,
            params=variables["params"],
            tx=optax.adamw(3e-4),
        )
        return state

    def train_step(self, x: Array, y: Array, state: TrainState):
        def loss_fn(params: FrozenVariableDict):
            y_pred = state.apply_fn({"params": params}, x)
            loss = optax.softmax_cross_entropy_with_integer_labels(y_pred, y).mean()
            return loss

        grads = jax.grad(loss_fn)(state.params)
        state = state.apply_gradients(grads=grads)
        return state



import json
from pathlib import Path
from typing import get_args
import click
from tabulate import tabulate
import importlib.util
import inspect

from ai_hexagon.test_suite import TestSuite
from ai_hexagon.tests import Test
from ai_hexagon.model import Model


@click.group()
def cli():
    pass


@cli.group()
def tests():
    pass


@tests.command()
@click.argument("test_name")
def show(test_name: str):
    tests = {t.get_test_name(): t for t in get_args(Test)}
    test = tests[test_name]
    print(f"Title: {test.__title__}")
    print(f"Description: {test.__description__}")
    print()
    print("Schema:")
    print(json.dumps(test.model_json_schema(), indent=4))


@tests.command()
def list():
    headers = ["Test Name", "Test Title", "Description"]
    table_data = []

    for test in get_args(Test):
        name = test.get_test_name()
        table_data.append([name, test.__title__, test.__description__])

    print(tabulate(table_data, headers=headers, tablefmt="simple_grid"))


@cli.group()
def suite():
    pass


@suite.command()
@click.argument("model_path", type=click.Path(exists=True))
@click.option(
    "--suite_path", type=click.Path(exists=True), default=Path("./results/suite.json")
)
@click.option("--save", is_flag=True)
def run(model_path: Path, suite_path: Path, save: bool):
    suite = TestSuite(**json.load(open(suite_path)))
    print(suite.model_dump_json(indent=4))

    spec = importlib.util.spec_from_file_location("model_module", model_path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load module from {model_path}")

    model_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(model_module)

    classes = [
        cls
        for name, cls in inspect.getmembers(model_module, inspect.isclass)
        if issubclass(cls, Model) and cls is not Model
    ]

    if not classes:
        print(f"No subclass of Model found in {model_path}")
        return

    if len(classes) > 1:
        print(
            f"Multiple subclasses of Model found in {model_path}: {[cls.__name__ for cls in classes]}"
        )
        return

    model_class = classes[0]
    result = suite.evaluate(model_class)
    json_str = json.dumps(result.model_dump(), indent=4)
    print(json_str)
    if save:
        with open(Path(model_path).with_suffix(".result.json"), "w") as f:
            f.write(json_str)


if __name__ == "__main__":
    cli()



from typing import Union

from ai_hexagon.tests.hash_map import HashMap
from ai_hexagon.tests.state_tracking import StateTracking


Test = Union[HashMap, StateTracking]

__all__ = ["Test", "HashMap", "StateTracking"]



import flax.linen as nn
import jax.numpy as jnp
from flax.typing import Array


class SinEmbedding(nn.Module):
    base_freq: int

    @nn.compact
    def __call__(self, x: Array) -> Array:
        positions = jnp.arange(x.shape[-2])
        div_term = jnp.linspace(0, 1, num=x.shape[-1])[None]
        div_term = self.base_freq**div_term

        emb = positions / div_term
        emb = jnp.concatenate(
            [jnp.sin(emb[:, ::2])[..., None], jnp.cos(emb[:, 1::2])[..., None]], axis=-1
        )
        emb = emb.reshape((*emb.shape[:-2], -1))
        return emb



import random
from typing import Tuple, Type, Literal

from ai_hexagon.model import Model
from ai_hexagon.test import BaseTest


class StateTracking(BaseTest):
    __title__ = "State Tracking"
    __description__ = "Tests model ability to manipulate and track state"

    name: Literal["state_tracking"] = "state_tracking"

    num_steps_range: Tuple[int, int] = (2, 128)
    state_size: int = 16

    def evalulate(self, model: Type[Model]) -> float:
        # TODO: implement the testing
        return random.random()



from typing import Dict, List, Optional, Type
from pydantic import BaseModel

from ai_hexagon.model import Model, ModelStats
from ai_hexagon.tests import Test, HashMap, StateTracking


class WeightedTest(BaseModel):
    weight: float
    test: Test


class Metric(BaseModel):
    name: str
    description: str
    tests: List[WeightedTest]


class Results(BaseModel):
    title: str
    description: str
    authors: Optional[List[str]]
    paper: Optional[str]
    metrics: Dict[str, float]
    model_stats: ModelStats

    model_config = {"protected_namespaces": ()}


class TestSuite(BaseModel):
    name: str
    description: str
    vocab_size: int
    sequence_length: int
    sequence_lengths: List[int]
    metrics: List[Metric]

    def evaluate(self, model: Type[Model]) -> Results:
        weighted_tests = sum([m.tests for m in self.metrics], [])
        tests = {wt.test for wt in weighted_tests}
        test_results = {}
        for t in tests:
            test_results[t] = t.evalulate(model)
        metrics = {}
        for m in self.metrics:
            metrics[m.name] = sum([wt.weight * test_results[wt.test] for wt in m.tests])
        model_stats = model.compute_stats(
            self.vocab_size, self.sequence_length, self.sequence_lengths
        )
        return Results(
            title=model.get_model_title(),
            description=model.__doc__,
            authors=model.__authors__,
            paper=model.__paper__,
            metrics=metrics,
            model_stats=model_stats,
        )


if __name__ == "__main__":
    memory_capacity = Metric(
        name="Memory Capacity",
        description="The ability of the model to store and recall information from the training data.",
        tests=[
            WeightedTest(
                weight=1.0,
                test=HashMap(
                    key_length=8,
                    value_length=64,
                    num_pairs_range=(32, 65536),
                    vocab_size=1024,
                ),
            )
        ],
    )
    state_management = Metric(
        name="State Management",
        description="The ability to maintain and manipulate an internal hidden state across a sequence of operations.",
        tests=[
            WeightedTest(
                weight=1.0,
                test=StateTracking(num_steps=(2, 128), state_size=16),
            ),
        ],
    )
    suite = TestSuite(
        name="General 1M",
        description="General test of model architecture performance",
        vocab_size=16000,
        sequence_length=8192,
        sequence_lengths=[32, 64, 128, 256, 512, 1024, 2048, 4096, 8192],
        metrics=[memory_capacity, state_management],
    )
    print(suite.model_dump_json(indent=4))
